<?php

namespace Kahanit\TableRateShipping\Model\ResourceModel;

// inheritance
use Magento\Framework\Model\ResourceModel\Db\AbstractDb;
// injection
use Magento\Framework\Model\ResourceModel\Db\Context;
use Magento\Framework\Filesystem;
use Magento\Directory\Model\CountryFactory;
use Magento\Directory\Model\RegionFactory;
use Kahanit\TableRateShipping\Model\CarrierTableRateFactory;
use Kahanit\TableRateShipping\Model\CarrierTableRateZoneFactory;
// other
use Magento\Framework\App\Filesystem\DirectoryList;
use Kahanit\TableRateShipping\Helper\TRSHelper;
use Symfony\Component\ExpressionLanguage\ExpressionLanguage;
use Kahanit\TableRateShipping\Helper\TRSExpressionFunctionProvider;
use Magento\Store\Model\Store;

class CarrierTableRate extends AbstractDb
{
    protected $fileSystem;

    protected $carrierTableRateZoneFactory;

    protected $countryFactory;

    protected $regionFactory;

    protected $carrierTableRateFactory;

    public static $id_group_static = 0;

    public function __construct(
        Context $context,
        Filesystem $fileSystem,
        CountryFactory $countryFactory,
        RegionFactory $regionFactory,
        CarrierTableRateFactory $carrierTableRateFactory,
        CarrierTableRateZoneFactory $carrierTableRateZoneFactory,
        $connectionName = null
    ) {
        $this->fileSystem = $fileSystem;
        $this->countryFactory = $countryFactory;
        $this->regionFactory = $regionFactory;
        $this->carrierTableRateFactory = $carrierTableRateFactory;
        $this->carrierTableRateZoneFactory = $carrierTableRateZoneFactory;

        parent::__construct($context, $connectionName);
    }

    protected function _construct()
    {
        $this->_init('carrier_table_rate', 'id_carrier_table_rate');
    }

    public function validateFields(\Kahanit\TableRateShipping\Model\CarrierTableRate $carrierTableRate)
    {
        /* validate zone */
        $zone = $this->carrierTableRateZoneFactory->create()->load($carrierTableRate->getData('id_zone'));
        if ($carrierTableRate->getData('id_zone') != 0 && $zone->isEmpty()) {
            return 'Invalid `Zone`';
        }

        /* validate country */
        $country = $this->countryFactory->create()->load($carrierTableRate->getData('id_country'));
        if ($carrierTableRate->getData('id_country') != '0' && $country->isEmpty()) {
            return 'Invalid `Country`';
        }

        /* validate state */
        $state = $this->regionFactory->create()->load($carrierTableRate->getData('id_state'));
        if ($carrierTableRate->getData('id_state') != 0 && $state->isEmpty()) {
            return 'Invalid `State`';
        }

        /* validate city */
        if (!preg_match('/^[^!<>;?=+@#"°{}_$%]*$/u', $carrierTableRate->getData('dest_city'))) {
            return 'Invalid `City`';
        }

        /* validate zip codes */
        $dest_zip = explode('-', $carrierTableRate->getData('dest_zip'));
        $dest_zip_from = (isset($dest_zip[0])) ? trim($dest_zip[0]) : '';
        $dest_zip_from = ($dest_zip_from === '*' || $dest_zip_from === '') ? '' : $dest_zip_from;
        $dest_zip_to = (isset($dest_zip[1])) ? trim($dest_zip[1]) : '';
        $dest_zip_to = ($dest_zip_to === '*' || $dest_zip_to === '') ? '' : $dest_zip_to;
        $dest_zip_to = ($dest_zip_to === '') ? $dest_zip_from : $dest_zip_to;
        if (strcasecmp($dest_zip_from, $dest_zip_to) === 0) {
            $carrierTableRate->setDestZip($dest_zip_from);
        } else {
            $carrierTableRate->setDestZip($dest_zip_from . '-' . $dest_zip_to);
        }

        /* validate zip/post code */
        if (!preg_match('/^[0-9a-zA-Z\-\s\*%_]*$/', $dest_zip_from) || !preg_match('/^[0-9a-zA-Z\-\s\*%_]*$/', $dest_zip_to)) {
            return 'Invalid `Zip/Post code`';
        }

        /* validate zip from and to range */
        if ($dest_zip_from != '' && $dest_zip_to != '' && $dest_zip_to < $dest_zip_from) {
            return '`Zip/Post code from` should be less than `Zip/Post code to`';
        }

        /* validate to cannot specify range if wildcard is used */
        if ($dest_zip_from != '' && $dest_zip_to != '' && strcasecmp($dest_zip_from, $dest_zip_to) != 0
            && ((strpos($dest_zip_from, '*') !== false || strpos($dest_zip_to, '*') !== false)
                || (strpos($dest_zip_from, '%') !== false || strpos($dest_zip_to, '%') !== false)
                || (strpos($dest_zip_from, '_') !== false || strpos($dest_zip_to, '_') !== false))
        ) {
            return 'You cannot specify range if wildcard is used in zip/post code';
        }

        /* validate condition value from and to range */
        $condition_error_message = [
            'condition_weight_from_less_than_condition_weight_to' => '`Weight from` should be less than `Weight to`',
            'condition_price_from_less_than_condition_price_to' => '`Price from` should be less than `Price to`',
            'condition_ptprice_from_less_than_condition_ptprice_to' => '`Price(pretax) from` should be less than `Price(pretax) to`',
            'condition_quantity_from_less_than_condition_quantity_to' => '`Quantity from` should be less than `Quantity to`',
            'condition_volume_from_less_than_condition_volume_to' => '`Volume from` should be less than `Volume to`'
        ];
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];
        foreach ($condition_types as $condition_type) {
            if ($carrierTableRate->getData('condition_' . $condition_type . '_to') < $carrierTableRate->getData('condition_' . $condition_type . '_from')) {
                return $condition_error_message['condition_' . $condition_type . '_from_less_than_condition_' . $condition_type . '_to'];
            }
        }

        /* validate symfony express language component installed */
        if (!class_exists('Symfony\\Component\\ExpressionLanguage\\ExpressionLanguage')) {
            return 'Symfony expression language component not found';
        }

        /* validate price */
        $vars = [
            'tw' => 1,
            'tp' => 1,
            'tptp' => 1,
            'tq' => 1,
            'tv' => 1,
            'ctwf' => 1,
            'ctwt' => 1,
            'ctwi' => 1,
            'ctpf' => 1,
            'ctpt' => 1,
            'ctpi' => 1,
            'ctptpf' => 1,
            'ctptpt' => 1,
            'ctptpi' => 1,
            'ctqf' => 1,
            'ctqt' => 1,
            'ctqi' => 1,
            'ctvf' => 1,
            'ctvt' => 1,
            'ctvi' => 1,
            'skip' => -1,
            'stop' => false
        ];
        $carrierTableRate->setData('price', str_replace('$', '', $carrierTableRate->getData('price'))); // for bc with old formula variables with $
        $language = new ExpressionLanguage();
        $language->registerProvider(new TRSExpressionFunctionProvider());
        try {
            $language->evaluate($carrierTableRate->getData('price'), $vars);
        } catch (\Exception $e) {
            return 'Invalid `Price`';
        }

        /* validate duplicate entry */
        $rule_exists = $this->ruleExists($carrierTableRate->getIdCarrier(), $carrierTableRate->getData(), $carrierTableRate->getIdShop());
        if ($rule_exists !== 0 && $rule_exists !== $carrierTableRate->getId()) {
            return 'Duplicate table rate found';
        }

        return (int)true;
    }

    public function ruleExists($id_carrier, $record, $id_shop)
    {
        $connection = $this->getConnection();
        $sql = $connection->select();
        $sql->from($this->getMainTable(), ['id_carrier_table_rate']);
        $sql->where(
            '   id_carrier = ' . (int)$id_carrier . '
            AND id_zone = ' . (int)$record['id_zone'] . '
            AND id_country = \'' . TRSHelper::addslashes($record['id_country']) . '\'
            AND id_state = ' . (int)$record['id_state'] . '
            AND dest_city = \'' . TRSHelper::addslashes($record['dest_city']) . '\'
            AND dest_zip = \'' . TRSHelper::addslashes($record['dest_zip']) . '\'
            AND condition_weight_from = ' . (float)$record['condition_weight_from'] . '
            AND condition_weight_to = ' . (float)$record['condition_weight_to'] . '
            AND condition_price_from = ' . (float)$record['condition_price_from'] . '
            AND condition_price_to = ' . (float)$record['condition_price_to'] . '
            AND condition_ptprice_from = ' . (float)$record['condition_ptprice_from'] . '
            AND condition_ptprice_to = ' . (float)$record['condition_ptprice_to'] . '
            AND condition_quantity_from = ' . (float)$record['condition_quantity_from'] . '
            AND condition_quantity_to = ' . (float)$record['condition_quantity_to'] . '
            AND condition_volume_from = ' . (float)$record['condition_volume_from'] . '
            AND condition_volume_to = ' . (float)$record['condition_volume_to'] . '
            AND price = \'' . TRSHelper::addslashes($record['price']) . '\'
            AND id_shop = ' . (int)$id_shop
        );

        return (int)$connection->fetchOne($sql);
    }

    public function getRuleGroups($id_carrier, $id_zone = null, $start = null, $length = null, $where = '', $id_shop)
    {
        $connection = $this->getConnection();

        // select query
        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            [
                'ct.id_zone',
                'z.name AS zone',
                'ct.id_group',
                'ct.id_country',
                'ct.id_state',
                'ct.dest_city',
                'ct.dest_zip',
                'MIN(ct.order) AS order'
            ]
        );
        $sql->joinLeft(['z' => $this->getTable('carrier_table_rate_zone')], 'z.id_zone = ct.id_zone AND z.id_shop = ' . (int)$id_shop, []);
        $sql->where('ct.id_carrier = ' . (int)$id_carrier
            . (($id_zone !== null) ? ' AND ct.id_zone = ' . (int)$id_zone : '')
            . (($id_shop !== false) ? ' AND ct.id_shop = ' . (int)$id_shop : ''));
        $sql->group(['ct.id_group', 'ct.dest_zip']);
        $sql->order(['MIN(ct.order) ASC']);

        // outer query
        $sql_outer = $connection->select();
        $sql_outer->from(
            ['st' => $sql],
            [
                'st.id_zone',
                'st.zone',
                'st.id_group',
                'st.id_country',
                'st.id_state',
                'st.dest_city',
                'GROUP_CONCAT(st.dest_zip ORDER BY st.order ASC SEPARATOR \',\') AS dest_zip'
            ]
        );
        $sql_outer->group(['st.id_group']);
        $sql_outer->order(['MIN(st.order) ASC']);
        if ($start !== null && $length !== null) {
            $sql_outer->limit($length, $start);
        }

        $results = $connection->fetchAll($sql_outer);

        // get country names from id
        $country_id = array_unique(array_column($results, 'id_country'));
        $country = [];
        if (count($country_id) > 0) {
            $country_collection = $this->countryFactory->create()->getCollection();
            $country_collection->addFieldToFilter('country_id', ['IN', $country_id]);
            foreach ($country_collection->getItems() as $key => $value) {
                $country[$key] = [
                    'iso_code' => $value->getData('iso2_code'),
                    'name' => $value->getName()
                ];
            }
        }

        // get state names from id
        $state_id = array_unique(array_column($results, 'id_state'));
        $state = [];
        if (count($state_id) > 0) {
            $state_collection = $this->regionFactory->create()->getCollection();
            $state_collection->addFieldToFilter('main_table.region_id', ['IN', $state_id]);
            foreach ($state_collection->getItems() as $key => $value) {
                $state[$key] = [
                    'iso_code' => $value->getData('code'),
                    'name' => $value->getName()
                ];
            }
        }

        foreach ($results as &$result) {
            $result['id_zone'] = ($result['id_zone'] == 0) ? '*' : $result['id_zone'];
            $result['zone'] = ($result['zone'] === '') ? '*' : $result['zone'];
            $result['id_country'] = ($result['id_country'] == '0') ? '*' : $result['id_country'];
            $result['c_iso_code'] = ($result['id_country'] == '0') ? '*' : ((isset($country[$result['id_country']])) ? $country[$result['id_country']]['iso_code'] : $result['id_country']);
            $result['country'] = ($result['id_country'] == '0') ? '*' : ((isset($country[$result['id_country']])) ? $country[$result['id_country']]['name'] : $result['id_country']);
            $result['id_state'] = ($result['id_state'] == 0) ? '*' : $result['id_state'];
            $result['s_iso_code'] = ($result['id_state'] == 0) ? '*' : ((isset($state[$result['id_state']])) ? $state[$result['id_state']]['iso_code'] : $result['id_state']);
            $result['state'] = ($result['id_state'] == 0) ? '*' : ((isset($state[$result['id_state']])) ? $state[$result['id_state']]['name'] : $result['id_state']);
            $result['dest_city'] = TRSHelper::stripslashes(($result['dest_city'] === '') ? '*' : $result['dest_city']);
            $result['dest_zip'] = TRSHelper::stripslashes(($result['dest_zip'] === '') ? '*' : $result['dest_zip']);
        }

        return $results;
    }

    public function getRuleGroupsNumRows($id_carrier, $id_zone = null, $where = '', $id_shop)
    {
        $connection = $this->getConnection();

        // sub query
        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            ['id_carrier_table_rate']
        );
        $sql->where(
            'ct.id_carrier = ' . (int)$id_carrier
            . (($id_zone !== null) ? ' AND ct.id_zone = ' . (int)$id_zone : '')
            . (($id_shop !== false) ? ' AND ct.id_shop = ' . (int)$id_shop : '')
        );
        $sql->group(['ct.id_group']);

        // count query
        $sql_count = 'SELECT COUNT(*) AS total FROM (' . $sql . ') AS tbl';

        return (int)$connection->fetchOne($sql_count);
    }

    public function getRules($id_carrier, $id_zone, $id_group, $start = null, $length = null, $where = '', $id_shop)
    {
        $connection = $this->getConnection();

        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            [
                'ct.*',
                'GROUP_CONCAT(ct.id_carrier_table_rate ORDER BY ct.order ASC SEPARATOR \',\') AS id_carrier_table_rate',
                'GROUP_CONCAT(ct.dest_zip ORDER BY ct.order ASC SEPARATOR \',\') AS dest_zip'
            ]
        );
        $sql->where(
            '   ct.id_carrier = ' . (int)$id_carrier . '
            AND ct.id_zone = ' . (int)$id_zone . '
            AND ct.id_group = ' . (int)$id_group . '
            AND ct.id_shop = ' . (int)$id_shop
        );
        $sql->group(
            [
                'ct.condition_weight_from',
                'ct.condition_weight_to',
                'ct.condition_price_from',
                'ct.condition_price_to',
                'ct.condition_ptprice_from',
                'ct.condition_ptprice_to',
                'ct.condition_quantity_from',
                'ct.condition_quantity_to',
                'ct.condition_volume_from',
                'ct.condition_volume_to',
                'ct.price',
                'ct.active'
            ]
        );
        $sql->order(['MIN(ct.order) ASC']);
        if ($start !== null && $length !== null) {
            $sql->limit($length, $start);
        }

        $results = $connection->fetchAll($sql);
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];
        foreach ($results as &$result) {
            $result['id_country'] = ($result['id_country'] == '0') ? '*' : $result['id_country'];
            $result['id_state'] = ($result['id_state'] == 0) ? '*' : $result['id_state'];
            $result['dest_city'] = TRSHelper::stripslashes(($result['dest_city'] === '') ? '*' : $result['dest_city']);
            $result['dest_zip'] = TRSHelper::stripslashes(($result['dest_zip'] === '') ? '*' : $result['dest_zip']);
            foreach ($condition_types as $condition_type) {
                $result['condition_' . $condition_type . '_from'] = ($result['condition_' . $condition_type . '_from'] == -1) ? '*' : $result['condition_' . $condition_type . '_from'];
                $result['condition_' . $condition_type . '_to'] = ($result['condition_' . $condition_type . '_to'] == -1) ? '*' : $result['condition_' . $condition_type . '_to'];
            }
            $result['price'] = TRSHelper::stripslashes($result['price']);
            $result['comment'] = TRSHelper::stripslashes($result['comment']);
        }

        return $results;
    }

    public function getRulesNumRows($id_carrier, $id_zone, $id_group, $where, $id_shop)
    {
        $connection = $this->getConnection();

        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            ['GROUP_CONCAT(ct.id_carrier_table_rate ORDER BY ct.order ASC SEPARATOR \',\') AS id_carrier_table_rate']
        );
        $sql->where(
            '   ct.id_carrier = ' . (int)$id_carrier . '
            AND ct.id_zone = ' . (int)$id_zone . '
            AND ct.id_group = ' . (int)$id_group . '
            AND ct.id_shop = ' . (int)$id_shop
        );
        $sql->group(
            [
                'ct.condition_weight_from',
                'ct.condition_weight_to',
                'ct.condition_price_from',
                'ct.condition_price_to',
                'ct.condition_ptprice_from',
                'ct.condition_ptprice_to',
                'ct.condition_quantity_from',
                'ct.condition_quantity_to',
                'ct.condition_volume_from',
                'ct.condition_volume_to',
                'ct.price',
                'ct.active'
            ]
        );
        $sql->order(['MIN(ct.order) ASC']);
        $sql_count = 'SELECT COUNT(id_carrier_table_rate) AS total FROM (' . $sql . ') AS tbl';

        return (int)$connection->fetchOne($sql_count);
    }

    public function saveRules($action, $id_carrier, $records, $id_shop)
    {
        $connection = $this->getConnection();
        $dir_var = $this->fileSystem->getDirectoryWrite(DirectoryList::VAR_DIR);
        if (!$dir_var->isFile('import/Kahanit/TableRateShipping/ImportStatus.json')) {
            $dir_var->writeFile('import/Kahanit/TableRateShipping/ImportStatus.json', '{"rowImported":0}');
        }

        $status = [
            'error' => 0,
            'success' => 0,
            'status' => '',
            'message' => []
        ];
        $fields = [
            'id_shop' => 0,
            'id_carrier' => 0,
            'id_zone' => 0,
            'id_group' => 0,
            'id_country' => '0',
            'id_state' => 0,
            'dest_city' => '',
            'dest_zip' => '',
            'condition_weight_from' => -1,
            'condition_weight_to' => -1,
            'condition_price_from' => -1,
            'condition_price_to' => -1,
            'condition_ptprice_from' => -1,
            'condition_ptprice_to' => -1,
            'condition_quantity_from' => -1,
            'condition_quantity_to' => -1,
            'condition_volume_from' => -1,
            'condition_volume_to' => -1,
            'price' => '',
            'comment' => '',
            'active' => 0,
            'order' => 0
        ];
        $fields_group = [
            'id_shop' => 0,
            'id_carrier' => 0,
            'id_zone' => 0,
            'id_country' => '0',
            'id_state' => 0,
            'dest_city' => ''
        ];
        $condition_types = [
            'weight',
            'price',
            'ptprice',
            'quantity',
            'volume'
        ];
        $insert_carrier_table_rates = [];
        $insert_carrier_table_rates_flat = [];
        $insert_carrier_table_groups_flat = [];

        if (is_array($records) && count($records) > 0) {
            foreach ($records as $record_index => &$record) {
                // assign shop and carrier
                $record = ['id_shop' => $id_shop, 'id_carrier' => $id_carrier] + $record;

                // prepare condition from and to values
                foreach ($condition_types as $condition_type) {
                    if (isset($record['condition_' . $condition_type])) {
                        if (is_numeric($record['condition_' . $condition_type])) {
                            $record['condition_' . $condition_type . '_from'] = (float)$record['condition_' . $condition_type];
                            $record['condition_' . $condition_type . '_to'] = (float)$record['condition_' . $condition_type];
                        } else {
                            $condition_values = explode('-', $record['condition_' . $condition_type]);
                            if (isset($condition_values[0]) && is_numeric($condition_values[0])
                                && isset($condition_values[1]) && is_numeric($condition_values[1])
                            ) {
                                $record['condition_' . $condition_type . '_from'] = (float)$condition_values[0];
                                $record['condition_' . $condition_type . '_to'] = (float)$condition_values[1];
                            }
                        }
                        unset($record['condition_' . $condition_type]);
                    }
                }

                // prepare record
                foreach ($fields as $field_key => $field_default) {
                    if (isset($record[$field_key])) {
                        $record[$field_key] = ($record[$field_key] === '*' || $record[$field_key] === '') ? $field_default : $record[$field_key];
                    } else {
                        $record[$field_key] = $field_default;
                    }
                }
                $record = array_merge($fields, $record);

                // prepare list of zipcodes
                if (isset($record['dest_zip'])) {
                    $record['dest_zip'] = (($record['dest_zip'] === '*' || $record['dest_zip'] === '') ? '' : $record['dest_zip']);
                } else {
                    $record['dest_zip'] = '';
                }
                $record['dest_zip'] = preg_replace('/[\r\n,]+/', '{{separator}}', $record['dest_zip']);
                $record['dest_zip'] = explode('{{separator}}', $record['dest_zip']);

                // get group rules
                if (!isset($record['id_group']) || $record['id_group'] == 0) {
                    $record_temp = $record;
                    $record_temp['dest_zip'] = implode(',', $record_temp['dest_zip']);
                    $insert_carrier_table_group_flat = TRSHelper::addslashes(implode(',', array_intersect_key($record_temp, $fields_group + ['dest_zip' => ''])));
                    if (count($insert_carrier_table_rates) > 0 && in_array($insert_carrier_table_group_flat, $insert_carrier_table_groups_flat)) {
                        $record['id_group'] = array_search($insert_carrier_table_group_flat, $insert_carrier_table_groups_flat);
                    } else {
                        $record['id_group'] = $this->getIdGroup($record['id_carrier'], $record['id_zone'], TRSHelper::addslashes($record['id_country']),
                            $record['id_state'], TRSHelper::addslashes($record['dest_city']), TRSHelper::addslashes(implode(',', $record['dest_zip'])), $record['id_shop']);
                    }
                }
                $rules = self::getRules($record['id_carrier'], $record['id_zone'], $record['id_group'], null, null, '', $record['id_shop']);
                $group_zips = [];
                if (is_array($rules) && count($rules) > 0) {
                    $group_zips = explode(',', reset($rules)['dest_zip']);
                }

                // prepare delete rules with `group zips` not in `dest zips sent`
                $delete_zips = [];
                if ($action == 'edit') {
                    $delete_zips = array_diff($group_zips, $record['dest_zip']);
                    if (count($delete_zips) > 0) {
                        $group_zips = array_diff($group_zips, $delete_zips);
                    }
                }

                // insert `all rules in database` for `dest zips sent`
                if (is_array($rules) && count($rules) > 0) {
                    foreach ($rules as $rule) {
                        $group_zips_temp = '';
                        foreach ($record['dest_zip'] as $dest_zip) {
                            // if dest zip is in database continue
                            if (in_array($dest_zip, $group_zips)) {
                                continue;
                            }

                            // setup object
                            $carrier_table_rate = $this->carrierTableRateFactory->create();
                            foreach ($record as $record_key => $record_value) {
                                if (array_key_exists($record_key, $fields)) {
                                    if (is_int($fields[$record_key])) {
                                        $carrier_table_rate->setData($record_key, (float)$record_value);
                                    } else {
                                        $carrier_table_rate->setData($record_key, TRSHelper::addslashes($record_value));
                                    }
                                }
                            }

                            // assign rule data
                            foreach ($condition_types as $condition_type) {
                                $carrier_table_rate->setData('condition_' . $condition_type . '_from',
                                    (($rule['condition_' . $condition_type . '_from'] === '*' || $rule['condition_' . $condition_type . '_from'] === '')
                                        ? -1 : $rule['condition_' . $condition_type . '_from']));
                                $carrier_table_rate->setData('condition_' . $condition_type . '_to',
                                    (($rule['condition_' . $condition_type . '_to'] === '*' || $rule['condition_' . $condition_type . '_to'] === '')
                                        ? -1 : $rule['condition_' . $condition_type . '_to']));
                            }
                            $carrier_table_rate->setData('price', TRSHelper::addslashes($rule['price']));
                            $carrier_table_rate->setData('comment', TRSHelper::addslashes($rule['comment']));
                            $carrier_table_rate->setData('active', (int)$rule['active']);

                            // assign dest zip
                            $carrier_table_rate->setData('dest_zip', TRSHelper::addslashes($dest_zip));

                            // validate object
                            $message = $this->validateFields($carrier_table_rate);

                            // if duplicate record found continue
                            if (strcasecmp($message, 'Duplicate table rate found') === 0) {
                                continue;
                            }

                            // save object of validated successfully
                            if ($message == 1) {
                                $insert_carrier_table_rate = array_intersect_key($carrier_table_rate->getData(), $fields);
                                $insert_carrier_table_rate_flat = implode(',', $insert_carrier_table_rate);
                                if (!in_array($insert_carrier_table_rate_flat, $insert_carrier_table_rates_flat)) {
                                    $insert_carrier_table_rates_flat[] = $insert_carrier_table_rate_flat;
                                    $group_zips_temp .= ',' . $carrier_table_rate->getData('dest_zip');
                                    $insert_carrier_table_group_flat = implode(',', array_intersect_key($carrier_table_rate->getData(), $fields_group));
                                    $insert_carrier_table_groups_flat[$insert_carrier_table_rate['id_group']] = $insert_carrier_table_group_flat . $group_zips_temp;
                                    $insert_carrier_table_rates[] = $insert_carrier_table_rate;
                                    $status['success']++;
                                }
                            } else {
                                $status['error']++;
                                $status['message'][] = $message;
                            }
                        }
                    }
                    $group_zips = array_unique(array_merge(array_diff($group_zips, $record['dest_zip']), $record['dest_zip']));
                } else {
                    $group_zips = $record['dest_zip'];
                }

                // insert `rule sent` for `database dest zips`
                if (is_array($group_zips) && count($group_zips) > 0) {
                    $group_zips_temp = '';
                    foreach ($group_zips as $group_zip) {
                        // if group exists and sent rule is dummy continue
                        if (is_array($rules) && count($rules) > 0 && strcasecmp($record['comment'], '{{dummy}}') === 0) {
                            continue;
                        }

                        // setup object
                        $carrier_table_rate = $this->carrierTableRateFactory->create();
                        foreach ($record as $record_key => $record_value) {
                            if (array_key_exists($record_key, $fields)) {
                                if (is_int($fields[$record_key])) {
                                    $carrier_table_rate->setData($record_key, (float)$record_value);
                                } else {
                                    $carrier_table_rate->setData($record_key, TRSHelper::addslashes($record_value));
                                }
                            }
                        }

                        // assign dest zip
                        $carrier_table_rate->setData('dest_zip', TRSHelper::addslashes($group_zip));

                        // validate object
                        $message = $this->validateFields($carrier_table_rate);

                        // if duplicate record found continue
                        if (strcasecmp($message, 'Duplicate table rate found') === 0) {
                            continue;
                        }

                        // save object of validated successfully
                        if ($message == 1) {
                            $insert_carrier_table_rate = array_intersect_key($carrier_table_rate->getData(), $fields);
                            $insert_carrier_table_rate_flat = implode(',', $insert_carrier_table_rate);
                            if (!in_array($insert_carrier_table_rate_flat, $insert_carrier_table_rates_flat)) {
                                $insert_carrier_table_rates_flat[] = $insert_carrier_table_rate_flat;
                                $group_zips_temp .= ',' . $carrier_table_rate->getData('dest_zip');
                                $insert_carrier_table_group_flat = implode(',', array_intersect_key($carrier_table_rate->getData(), $fields_group));
                                $insert_carrier_table_groups_flat[$insert_carrier_table_rate['id_group']] = $insert_carrier_table_group_flat . $group_zips_temp;
                                $insert_carrier_table_rates[] = $insert_carrier_table_rate;
                                $status['success']++;
                            }
                        } else {
                            $status['error']++;
                            $status['message'][] = $message;
                        }
                    }
                }

                // execute delete rules with `group zips` not in `dest zips sent`
                if ($action == 'edit' && $status['error'] <= 0 && count($delete_zips) > 0) {
                    $connection->delete(
                        $this->getMainTable(),
                        '   id_carrier = ' . (int)$record['id_carrier'] . '
                        AND id_zone = ' . (int)$record['id_zone'] . '
                        AND id_group = ' . (int)$record['id_group'] . '
                        AND dest_zip IN (\'' . implode("','", $delete_zips) . '\')
                        AND id_shop = ' . (int)$id_shop
                    );
                }

                // write import status
                $dir_var->writeFile('import/Kahanit/TableRateShipping/ImportStatus.json', '{"rowImported":' . ((int)$record_index + 1) . '}');
            }
        }

        if ($status['error'] > 0) {
            $status['status'] = 'danger';
            $status['message'] = implode(', ', array_unique($status['message']));

            return $status;
        } else {
            // insert records
            if (count($insert_carrier_table_rates) > 0) {
                $connection->insertMultiple($this->getMainTable(), $insert_carrier_table_rates);
                $connection->query('
                    UPDATE `' . $this->getMainTable() . '`
                    SET `order` = `id_carrier_table_rate` WHERE `order` = 0
                ');
            }

            // sort zones
            $zones = $this->getZonesOrder($id_carrier, $id_shop);
            $this->updateOrder('zone', $id_carrier, null, $zones, $id_shop);

            // sort rule groups
            foreach ($zones as $id_zone) {
                $results = $this->getRuleGroups($id_carrier, $id_zone, null, null, '', $id_shop);
                $this->updateOrder('rulegroup', $id_carrier, $id_zone, array_column($results, 'id_group'), $id_shop);
            }

            // sort rules only on edit
            if ($action == 'edit' && isset($record['id_zone']) && isset($record['id_group']) && isset($group_zips)) {
                $results = $this->getRules($id_carrier, $record['id_zone'], $record['id_group'], null, null, '', $id_shop);
                $rules = [];
                foreach ($results as $result) {
                    $id_carrier_table_rate_explode = explode(',', $result['id_carrier_table_rate']);
                    $dest_zip_explode = explode(',', $result['dest_zip']);
                    $id_carrier_table_rate_explode_sorted = [];
                    foreach ($group_zips as $group_zip) {
                        $id_carrier_table_rate_explode_sorted[]
                            = $id_carrier_table_rate_explode[array_search($group_zip, $dest_zip_explode)];
                    }
                    $rules[] = implode(',', $id_carrier_table_rate_explode_sorted);
                }
                $this->updateOrder('rule', null, null, $rules, $id_shop);
            }

            // set success status
            $status['status'] = 'success';
            $status['message'] = 'Records saved successfully';

            return $status;
        }
    }

    public function import($file, $csv_separator = ',', $id_carrier, $id_shop)
    {
        $first_record = true;
        $valid_columns = [
            'zone',
            'country',
            'state',
            'city',
            'zip',
            'condition_weight',
            'condition_price',
            'condition_ptprice',
            'condition_quantity',
            'condition_volume',
            'price',
            'active',
            'comment'
        ];
        $column_numbers = [];
        $records = [];

        if (($handle = fopen($file, 'r')) !== false) {
            while (($data = fgetcsv($handle, 0, $csv_separator)) !== false) {
                if ($first_record) {
                    foreach ($data as $key => $value) {
                        if (in_array($value, $valid_columns)) {
                            $column_numbers[$value] = $key;
                        }
                    }

                    if (count($column_numbers) != count($valid_columns)) {
                        return [
                            'status' => 'danger',
                            'message' => 'CSV not properly formatted'
                        ];
                    }

                    $first_record = false;
                    continue;
                }

                $id_zone = ($data[$column_numbers['zone']] === '*' || $data[$column_numbers['zone']] === '')
                    ? 0 : $this->carrierTableRateZoneFactory->create()->getResource()->getIdByName($data[$column_numbers['zone']]);
                $id_country = ($data[$column_numbers['country']] === '*' || $data[$column_numbers['country']] === '')
                    ? '0' : $this->getCountryIdByIso($data[$column_numbers['country']]);
                $id_state = ($data[$column_numbers['state']] === '*' || $data[$column_numbers['state']] === '')
                    ? 0 : $this->getStateIdByIso($data[$column_numbers['state']], $id_country);

                $records[] = [
                    'id_zone' => $id_zone,
                    'id_country' => $id_country,
                    'id_state' => $id_state,
                    'dest_city' => ($data[$column_numbers['city']] === '*' || $data[$column_numbers['city']] === '')
                        ? '' : $data[$column_numbers['city']],
                    'dest_zip' => ($data[$column_numbers['zip']] === '*' || $data[$column_numbers['zip']] === '')
                        ? '' : preg_replace('/[\r\n,]+/', '{{separator}}', $data[$column_numbers['zip']]),
                    'condition_weight' => ($data[$column_numbers['condition_weight']] === '*' || $data[$column_numbers['condition_weight']] === '')
                        ? -1 : $data[$column_numbers['condition_weight']],
                    'condition_price' => ($data[$column_numbers['condition_price']] === '*' || $data[$column_numbers['condition_price']] === '')
                        ? -1 : $data[$column_numbers['condition_price']],
                    'condition_ptprice' => ($data[$column_numbers['condition_ptprice']] === '*' || $data[$column_numbers['condition_ptprice']] === '')
                        ? -1 : $data[$column_numbers['condition_ptprice']],
                    'condition_quantity' => ($data[$column_numbers['condition_quantity']] === '*' || $data[$column_numbers['condition_quantity']] === '')
                        ? -1 : $data[$column_numbers['condition_quantity']],
                    'condition_volume' => ($data[$column_numbers['condition_volume']] === '*' || $data[$column_numbers['condition_volume']] === '')
                        ? -1 : $data[$column_numbers['condition_volume']],
                    'price' => $data[$column_numbers['price']],
                    'active' => $data[$column_numbers['active']],
                    'comment' => $data[$column_numbers['comment']]
                ];
            }

            fclose($handle);
        }

        return $this->saveRules('add', $id_carrier, $records, $id_shop);
    }

    public function export($id_carrier, $csv_separator = ',', $id_shop)
    {
        $collection = [];
        $rule_groups = $this->getRuleGroups($id_carrier, null, null, null, '', $id_shop);
        foreach ($rule_groups as $rule_group) {
            $rules = $this->getRules($id_carrier, $rule_group['id_zone'], $rule_group['id_group'], null, null, '', $id_shop);

            foreach ($rules as $rule) {
                $weight = ($rule['condition_weight_from'] == $rule['condition_weight_to']) ? $rule['condition_weight_from'] : $rule['condition_weight_from'] . '-' . $rule['condition_weight_to'];
                $price = ($rule['condition_price_from'] == $rule['condition_price_to']) ? $rule['condition_price_from'] : $rule['condition_price_from'] . '-' . $rule['condition_price_to'];
                $ptprice = ($rule['condition_ptprice_from'] == $rule['condition_ptprice_to']) ? $rule['condition_ptprice_from'] : $rule['condition_ptprice_from'] . '-' . $rule['condition_ptprice_to'];
                $quantity = ($rule['condition_quantity_from'] == $rule['condition_quantity_to']) ? $rule['condition_quantity_from'] : $rule['condition_quantity_from'] . '-' . $rule['condition_quantity_to'];
                $volume = ($rule['condition_volume_from'] == $rule['condition_volume_to']) ? $rule['condition_volume_from'] : $rule['condition_volume_from'] . '-' . $rule['condition_volume_to'];

                $collection[] = [
                    'zone' => ($rule_group['zone'] === '') ? '*' : $rule_group['zone'],
                    'country' => ($rule_group['c_iso_code'] === '') ? '*' : $rule_group['c_iso_code'],
                    'state' => ($rule_group['s_iso_code'] === '') ? '*' : $rule_group['s_iso_code'],
                    'city' => ($rule_group['dest_city'] === '') ? '*' : $rule_group['dest_city'],
                    'zip' => ($rule_group['dest_zip'] === '') ? '*' : implode(PHP_EOL, explode(',', $rule_group['dest_zip'])),
                    'condition_weight' => ($weight == -1) ? '*' : $weight,
                    'condition_price' => ($price == -1) ? '*' : $price,
                    'condition_ptprice' => ($ptprice == -1) ? '*' : $ptprice,
                    'condition_quantity' => ($quantity == -1) ? '*' : $quantity,
                    'condition_volume' => ($volume == -1) ? '*' : $volume,
                    'price' => $rule['price'],
                    'comment' => $rule['comment'],
                    'active' => $rule['active']
                ];
            }
        }

        if (count($collection) > 0) {
            $csv = fopen("php://output", 'w');
            fputcsv($csv, array_keys(reset($collection)));
            foreach ($collection as $row) {
                fputcsv($csv, $row, $csv_separator);
            }
            fclose($csv);
            header('Content-type: text/csv');
            header('Content-Type: application/force-download; charset=UTF-8');
            header('Cache-Control: no-store, no-cache');
            header('Content-disposition: attachment; filename="export.csv"');
        } else {
            echo 'No records found';
        }

        die();
    }

    public function updateCountryStateCity($update, $id_carrier, $id_zone, $id_group, $value, $id_shop)
    {
        if ($update !== false && in_array($update, ['id_country', 'id_state', 'dest_city']) && $id_carrier !== 0 && $id_zone !== -1) {
            if ($update == 'id_country') {
                $value = TRSHelper::addslashes(($value === '*' || $value === '') ? '0' : $value);
            } elseif ($update == 'id_state') {
                $value = (int)(($value === '*' || $value === '') ? 0 : $value);
            } else {
                $value = TRSHelper::addslashes(($value === '*' || $value === '') ? '' : $value);
            }

            // validate group already exists
            $rules = $this->getRules($id_carrier, $id_zone, $id_group, null, null, '', $id_shop);
            foreach ($rules as $rule) {
                $id_carrier_table_rate_explode = array_map('intval', explode(',', $rule['id_carrier_table_rate']));
                foreach ($id_carrier_table_rate_explode as $id_carrier_table_rate_explode_item) {
                    $carrier_table_rate = $this->carrierTableRateFactory->create()->load($id_carrier_table_rate_explode_item);
                    $carrier_table_rate->setData($update, $value);
                    if ($update == 'id_country') {
                        $carrier_table_rate->setData('id_state', 0);
                    }
                    $message = $this->validateFields($carrier_table_rate);
                    if ($message != 1) {
                        return [
                            'status' => 'danger',
                            'message' => $message
                        ];
                    }
                }
            }

            // update values
            $update_values = [$update => $value];
            if ($update == 'id_country') {
                $update_values['id_state'] = 0;
            }
            $connection = $this->getConnection();
            $connection->update(
                $this->getMainTable(),
                $update_values,
                '    id_carrier = ' . (int)$id_carrier . '
                AND id_zone = ' . (int)$id_zone . '
                AND id_group = ' . (int)$id_group . '
                AND id_shop = ' . (int)$id_shop
            );

            if ((int)$connection->fetchOne('SELECT ROW_COUNT()') > 0) {
                $rules = $this->getRules($id_carrier, $id_zone, $id_group, 0, 1, '', $id_shop);
                if (count($rules) > 0) {
                    $country = $this->countryFactory->create()->load($rules[0]['id_country']);
                    $state = $this->regionFactory->create()->load($rules[0]['id_state']);
                    return [
                        'status' => 'success',
                        'id_shop' => $id_shop,
                        'id_carrier' => $id_carrier,
                        'id_zone' => $id_zone,
                        'id_country' => ($rules[0]['id_country'] == '0') ? '*' : $rules[0]['id_country'],
                        'country' => ($country->getName() && $country->getData('iso2_code')) ? $country->getName() . ' (' . $country->getData('iso2_code') . ')' : '*',
                        'id_state' => ($rules[0]['id_state'] == 0) ? '*' : $rules[0]['id_state'],
                        'state' => ($state->getName() && $state->getData('code')) ? $state->getName() . ' (' . $state->getData('code') . ')' : '*',
                        'dest_city' => TRSHelper::stripslashes(($rules[0]['dest_city'] === '') ? '*' : $rules[0]['dest_city']),
                        'dest_zip' => TRSHelper::stripslashes(($rules[0]['dest_zip'] === '') ? '*' : $rules[0]['dest_zip']),
                        'message' => 'Record updated successfully'
                    ];
                }
            }
        }

        return [
            'status' => 'danger',
            'message' => 'Error updating record'
        ];
    }

    public function updateConditionPriceComment($update, $id_carrier_table_rate, $value)
    {
        $fields = [
            'condition_weight',
            'condition_price',
            'condition_ptprice',
            'condition_quantity',
            'condition_volume',
            'price',
            'comment'
        ];
        $condition_types = ['weight', 'price', 'ptprice', 'quantity', 'volume'];

        if ($update !== false && in_array($update, $fields) && $id_carrier_table_rate !== false) {
            if ($update == 'price' || $update == 'comment') {
                $update_values = [
                    $update => TRSHelper::addslashes($value)
                ];
            } else {
                if (is_numeric($value)) {
                    $value_from = $value_to = $value;
                } else {
                    $value_range = explode('-', $value);
                    if (isset($value_range[0]) && is_numeric($value_range[0]) && isset($value_range[1]) && is_numeric($value_range[1])) {
                        $value_from = $value_range[0];
                        $value_to = $value_range[1];
                    } else {
                        $value_from = $value_to = -1;
                    }
                }
                $update_values = [
                    $update . '_from' => (float)(($value_from === '*' || $value_from === '') ? -1 : $value_from),
                    $update . '_to' => (float)(($value_to === '*' || $value_to === '') ? -1 : $value_to)
                ];
            }

            // validate records
            $id_carrier_table_rate_explode = array_map('intval', explode(',', $id_carrier_table_rate));
            foreach ($id_carrier_table_rate_explode as $id_carrier_table_rate_explode_item) {
                $carrier_table_rate = $this->carrierTableRateFactory->create()->load($id_carrier_table_rate_explode_item);
                foreach ($update_values as $update_value_key => $update_value_value) {
                    $carrier_table_rate->setData($update_value_key, $update_value_value);
                }
                $message = $this->validateFields($carrier_table_rate);
                if ($message != 1) {
                    return [
                        'status' => 'danger',
                        'message' => $message
                    ];
                }
            }

            // update values
            $connection = $this->getConnection();
            $connection->update(
                $this->getMainTable(),
                $update_values,
                'id_carrier_table_rate IN (' . implode(',', array_map('intval', explode(',', $id_carrier_table_rate))) . ')'
            );

            if ((int)$connection->fetchOne('SELECT ROW_COUNT()') > 0) {
                $id_carrier_table_rate = array_map('intval', explode(',', $id_carrier_table_rate));
                if (isset($id_carrier_table_rate[0])) {
                    $carrier_table_rate = $this->carrierTableRateFactory->create()->load((int)$id_carrier_table_rate[0]);
                    $status = [
                        'status' => 'success',
                        'price' => TRSHelper::stripslashes($carrier_table_rate->getPrice()),
                        'comment' => TRSHelper::stripslashes($carrier_table_rate->getComment()),
                        'message' => 'Record updated successfully'
                    ];
                    foreach ($condition_types as $condition_type) {
                        $status['condition_' . $condition_type . '_from'] = ($carrier_table_rate->getData('condition_' . $condition_type . '_from') == -1)
                            ? '*' : $carrier_table_rate->getData('condition_' . $condition_type . '_from');
                        $status['condition_' . $condition_type . '_to'] = ($carrier_table_rate->getData('condition_' . $condition_type . '_to') == -1)
                            ? '*' : $carrier_table_rate->getData('condition_' . $condition_type . '_to');
                    }
                    return $status;
                }
            }
        }

        return [
            'status' => 'danger',
            'message' => 'Error updating record'
        ];
    }

    public function updateStatus($id_carrier_table_rate, $active)
    {
        if ($id_carrier_table_rate !== false) {
            $this->getConnection()->update(
                $this->getMainTable(),
                ['active' => (int)$active],
                'id_carrier_table_rate IN (' . implode(',', array_map('intval', explode(',', $id_carrier_table_rate))) . ')'
            );

            return [
                'status' => 'success',
                'message' => 'Status updated successfully',
                'id_carrier_table_rate' => $id_carrier_table_rate,
                'active' => $active
            ];
        }

        return [
            'status' => 'danger',
            'message' => 'Status update failed',
            'id_carrier_table_rate' => $id_carrier_table_rate,
            'active' => $active
        ];
    }

    public function updateOrder($entity = '', $id_carrier = null, $id_zone = null, $order = [], $id_shop)
    {
        $connection = $this->getConnection();

        if ($entity !== '' && $order !== []) {
            switch ($entity) {
                case 'zone':
                    if ($id_carrier !== 0) {
                        $order_ids = [];
                        foreach ($order as $order_item) {
                            $order_ids[] = (int)preg_replace('/^zone-/', '', $order_item);
                        }
                        $sql_rank = 'SET @rank := 0';
                        $sql = 'UPDATE `' . $this->getMainTable() . '`
                            SET `order` = @rank := @rank + 1
                            WHERE `id_carrier` = ' . (int)$id_carrier . ' AND `id_shop` = ' . (int)$id_shop . '
                            ORDER BY FIELD(`id_zone`, ' . implode(',', $order_ids) . '), `order` ASC';
                        $connection->query($sql_rank);
                        $connection->query($sql);
                    }
                    break;
                case 'rulegroup':
                    if ($id_carrier !== 0 && $id_zone !== -1) {
                        $order = array_map('intval', $order);

                        // get min order rank
                        $sql = $connection->select();
                        $sql->from($this->getMainTable(), ['MIN(`order`) AS `minorder`']);
                        $sql->where('`id_group` IN (' . implode(',', $order) . ')');
                        $minorder = (int)$connection->fetchOne($sql);
                        if (!$minorder) {
                            $minorder = 1;
                        }
                        $minorder--;

                        // update order
                        $sql_rank = 'SET @rank := ' . (int)$minorder;
                        $sql = 'UPDATE `' . $this->getMainTable() . '`
                            SET `order` = @rank := @rank + 1
                            WHERE
                                    `id_carrier` = ' . (int)$id_carrier . '
                                AND `id_zone` = ' . (int)$id_zone . '
                                AND `id_group` IN (' . implode(',', $order) . ')
                                AND `id_shop` = ' . (int)$id_shop . '
                            ORDER BY FIELD(`id_group`, ' . implode(',', $order) . '), `order` ASC';
                        $connection->query($sql_rank);
                        $connection->query($sql);
                    }
                    break;
                case 'rule':
                    $order_item_count = count(explode(',', reset($order)));
                    $order_temp = [];
                    for ($i = 0; $i < $order_item_count; $i++) {
                        foreach ($order as $order_item) {
                            $order_item_explode = array_map('intval', explode(',', $order_item));
                            $order_temp[] = $order_item_explode[$i];
                        }
                    }
                    $id_carrier_table_rate = implode(',', $order_temp);

                    // get min order rank
                    $sql = $connection->select();
                    $sql->from($this->getMainTable(), ['MIN(`order`) AS `minorder`']);
                    $sql->where('`id_carrier_table_rate` IN (' . $id_carrier_table_rate . ')');
                    $minorder = (int)$connection->fetchOne($sql);
                    if (!$minorder) {
                        $minorder = 1;
                    }
                    $minorder--;

                    // update order
                    $sql_rank = 'SET @rank := ' . (int)$minorder;
                    $sql = 'UPDATE `' . $this->getMainTable() . '`
                        SET `order` = @rank := @rank + 1
                        WHERE `id_carrier_table_rate` IN (' . $id_carrier_table_rate . ')
                        ORDER BY FIELD(`id_carrier_table_rate`, ' . $id_carrier_table_rate . '), `order` ASC';
                    $connection->query($sql_rank);
                    $connection->query($sql);
                    break;
            }
        }

        return [
            'status' => 'success',
            'message' => 'Order saved successfully'
        ];
    }

    public function deleteRules($entity, $type, $id_carrier, $data, $id_zone, $id_carrier_table_rate, $id_shop)
    {
        $connection = $this->getConnection();

        if ($entity == 'rule') {
            if ($id_carrier_table_rate !== false && ($type == 'single' || $type == 'selected')) {
                $id_carrier_table_rate = array_map('intval', explode(',', $id_carrier_table_rate));
                $id_carrier_table_rate = implode(',', $id_carrier_table_rate);
                $connection->delete($this->getMainTable(),
                    'id_carrier_table_rate IN (' . $id_carrier_table_rate . ')');
            }
        } elseif ($entity == 'rulegroup') {
            if ($data !== false && ($type == 'single' || $type == 'selected')) {
                foreach ($data as $line) {
                    $connection->delete($this->getMainTable(),
                        '   id_carrier = ' . (int)$id_carrier . '
                        AND id_zone = ' . (int)$line['id_zone'] . '
                        AND id_group = ' . (int)$line['id_group'] . '
                        AND id_shop = ' . (int)$id_shop);
                }
            } elseif ($id_zone !== -1 && $type == 'all') {
                $connection->delete($this->getMainTable(),
                    'id_zone = ' . (int)$id_zone . ' AND id_shop = ' . (int)$id_shop);
            }
        }

        return [
            'status' => 'success',
            'message' => 'Table rates deleted successfully'
        ];
    }

    public function deleteRulesByCarrier($id_carrier)
    {
        $this->getConnection()->delete($this->getMainTable(), 'id_carrier = ' . (int)$id_carrier);
    }

    public function deleteRulesByZone($id_zone)
    {
        $this->getConnection()->delete($this->getMainTable(), 'id_zone = ' . (int)$id_zone);
    }

    public function getIdGroup($id_carrier, $id_zone, $id_country, $id_state, $dest_city, $dest_zip, $id_shop)
    {
        $connection = $this->getConnection();

        // select query
        $sql = $connection->select();
        $sql->from(
            ['ct' => $this->getMainTable()],
            [
                'ct.id_group',
                'ct.dest_zip',
                'MIN(ct.order) AS order'
            ]
        );
        $sql->where(
            '   id_carrier = ' . (int)$id_carrier . '
            AND id_zone = ' . (int)$id_zone . '
            AND id_country = \'' . TRSHelper::addslashes($id_country) . '\'
            AND id_state = ' . (int)$id_state . '
            AND dest_city = \'' . TRSHelper::addslashes($dest_city) . '\'
            AND id_shop = ' . (int)$id_shop
        );
        $sql->group(['ct.id_group', 'ct.dest_zip']);
        $sql->order(['MIN(ct.order) ASC']);

        // final query
        $sql_final = $connection->select();
        $sql_final->from(
            ['st' => $sql],
            ['st.id_group']
        );
        $sql_final->group(['st.id_group']);
        $sql_final->having('GROUP_CONCAT(st.dest_zip ORDER BY st.order ASC SEPARATOR \',\') = \'' . TRSHelper::addslashes($dest_zip) . '\'');
        $sql_final->order(['MIN(st.order) ASC']);
        $id_group = (int)$connection->fetchOne($sql_final);

        if ($id_group) {
            return $id_group;
        }

        if (self::$id_group_static == 0) {
            $sql = $connection->select();
            $sql->from($this->getMainTable(), ['MAX(id_group) AS id_group']);

            self::$id_group_static = (int)$connection->fetchOne($sql) + 1;
        } else {
            self::$id_group_static++;
        }

        return self::$id_group_static;
    }

    public function getZonesOrder($id_carrier, $id_shop)
    {
        $connection = $this->getConnection();
        $order = [];

        if ($id_carrier !== 0) {
            $sql = $connection->select();
            $sql->distinct();
            $sql->from($this->getMainTable(), ['id_zone']);
            $sql->where('id_carrier = ' . (int)$id_carrier . ' AND id_shop = ' . (int)$id_shop);
            $sql->order(['order ASC']);
            $results = $connection->fetchAll($sql);
            foreach ($results as $result) {
                $order[] = $result['id_zone'];
            }
        }

        return $order;
    }

    public function getCountries($search)
    {
        $collection = $this->countryFactory->create()->getCollection();

        $results = [];
        foreach ($collection->getItems() as $key => $value) {
            $name = $value->getName();
            if ($search === '' || stristr($name, $search) !== false) {
                $results[] = [
                    'id' => $value->getId(),
                    'text' => $name
                ];
            }
        }

        return $results;
    }

    public function getStates($search, $id_country)
    {
        $collection = $this->regionFactory->create()->getCollection()
            ->addFieldToFilter('country_id', ['=' => $id_country])
            ->addFieldToFilter('name', ['like' => '%' . TRSHelper::addslashes($search, true) . '%']);

        $results = [];
        foreach ($collection->getItems() as $key => $value) {
            $results[] = [
                'id' => $value->getId(),
                'text' => $value->getName()
            ];
        }

        return $results;
    }

    public function getCountryIdByIso($iso)
    {
        $collection = $this->countryFactory->create()->getCollection()
            ->addFieldToFilter('iso2_code', ['=', $iso]);
        $country = $collection->getFirstItem();

        return (int)$country->getId();
    }

    public function getStateIdByIso($iso, $id_country)
    {
        $collection = $this->regionFactory->create()->getCollection()
            ->addFieldToFilter('code', ['=', $iso])
            ->addFieldToFilter('country_id', ['=', $id_country]);
        $state = $collection->getFirstItem();

        return (int)$state->getId();
    }

    public function getShippingPrice($id_carrier, $id_country, $id_state, $dest_city, $dest_zip, $vars, $id_shop)
    {
        if (!class_exists('Symfony\\Component\\ExpressionLanguage\\ExpressionLanguage')) {
            return false;
        }

        $dest_zip = str_replace('-', '', str_replace(' ', '', $dest_zip));

        $connection = $this->getConnection();
        $sql = $connection->select();
        $sql->from($this->getMainTable(), ['*', 'SUBSTRING_INDEX(dest_zip, "-", 1) AS dest_zip_from', 'SUBSTRING_INDEX(dest_zip, "-", -1) AS dest_zip_to']);
        $sql->where('
                (id_shop    = ' . (int)$id_shop . '                                OR id_shop = ' . Store::DEFAULT_STORE_ID . ')
            AND id_carrier  = ' . (int)$id_carrier . '
            AND (id_country = \'' . TRSHelper::addslashes($id_country, true) . '\' OR id_country = \'0\')
            AND (id_state   = ' . (int)$id_state . '                               OR id_state   = 0)
            AND (dest_city  = \'' . TRSHelper::addslashes($dest_city, true) . '\'  OR dest_city  = \'\')
            AND (' . (float)$vars['tw'] . '   BETWEEN condition_weight_from        AND condition_weight_to   OR (condition_weight_from   < 0 AND condition_weight_to   < 0))
            AND (' . (float)$vars['tp'] . '   BETWEEN condition_price_from         AND condition_price_to    OR (condition_price_from    < 0 AND condition_price_to    < 0))
            AND (' . (float)$vars['tptp'] . ' BETWEEN condition_ptprice_from       AND condition_ptprice_to  OR (condition_ptprice_from  < 0 AND condition_ptprice_to  < 0))
            AND (' . (float)$vars['tq'] . '   BETWEEN condition_quantity_from      AND condition_quantity_to OR (condition_quantity_from < 0 AND condition_quantity_to < 0))
            AND (' . (float)$vars['tv'] . '   BETWEEN condition_volume_from        AND condition_volume_to   OR (condition_volume_from   < 0 AND condition_volume_to   < 0))
            AND active = 1
        ');
        $sql->having('
            (
                dest_zip_from = ""
                AND dest_zip_to = ""
            )
            OR (
                dest_zip_from = dest_zip_to
                AND \'' . TRSHelper::addslashes($dest_zip, true) . '\' LIKE REPLACE(REPLACE(REPLACE(dest_zip_to, "*", "%"), " ", ""), "-", "")
            )
            OR (
                dest_zip_from = dest_zip_to
                AND \'' . TRSHelper::addslashes($dest_zip, true) . '\' LIKE REPLACE(REPLACE(REPLACE(dest_zip_from, "*", "%"), " ", ""), "-", "")
            )
            OR (
                dest_zip_from != ""
                AND dest_zip_to != ""
                AND REPLACE(REPLACE(dest_zip_from, " ", ""), "-", "") <= \'' . TRSHelper::addslashes($dest_zip, true) . '\'
                AND REPLACE(REPLACE(dest_zip_to, " ", ""), "-", "") >= \'' . TRSHelper::addslashes($dest_zip, true) . '\'
            )
        ');
        $sql->order(['FIELD(`id_shop`, ' . $id_shop . ', ' . Store::DEFAULT_STORE_ID . ')', 'order ASC']);
        $results = $connection->fetchAll($sql);

        $language = new ExpressionLanguage();
        $language->registerProvider(new TRSExpressionFunctionProvider());
        $shipping_price = false;

        foreach ($results as $result) {
            $vars_temp = $vars;
            $vars_temp = array_merge($vars_temp, [
                'ctwf' => $result['condition_weight_from'],
                'ctwt' => $result['condition_weight_to'],
                'ctwi' => $vars_temp['tw'] - $result['condition_weight_from'],
                'ctpf' => $result['condition_price_from'],
                'ctpt' => $result['condition_price_to'],
                'ctpi' => $vars_temp['tp'] - $result['condition_price_from'],
                'ctptpf' => $result['condition_ptprice_from'],
                'ctptpt' => $result['condition_ptprice_to'],
                'ctptpi' => $vars_temp['tptp'] - $result['condition_ptprice_from'],
                'ctqf' => $result['condition_quantity_from'],
                'ctqt' => $result['condition_quantity_to'],
                'ctqi' => $vars_temp['tq'] - $result['condition_quantity_from'],
                'ctvf' => $result['condition_volume_from'],
                'ctvt' => $result['condition_volume_to'],
                'ctvi' => $vars_temp['tv'] - $result['condition_volume_from']
            ]);

            $result['price'] = str_replace('$', '', $result['price']); // for bc with old formula variables with $

            try {
                $shipping_price = $language->evaluate($result['price'], $vars_temp);
            } catch (\Exception $e) {
                $shipping_price = false;
            }

            if ($shipping_price < 0) {
                continue;
            } else {
                break;
            }
        }

        return $shipping_price;
    }
}
